import json

from django.contrib.auth.decorators import login_required
from django.apps import apps
from django.http import HttpResponse
from .filter import Filter


@login_required
def params(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    return HttpResponse(Filter(Model, request.GET).query_params())


@login_required
def set_filter_input(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    lookup = request.GET.get('lookup')
    return HttpResponse(Filter(Model, request.GET).query_input(lookup))


@login_required
def filter_add_query(request, model: str):
    app_label, model_name = model.split('.')
    Model = apps.get_model(app_label, model_name)
    query = json.loads(request.GET.get('q', '[]'))
    lookup = request.GET.get('filter_lookup')
    value = request.GET.get('filter_input')
    query.append({lookup: value})
    query = json.dumps(query)
    query_dict = request.GET.copy()
    query_dict.update(q=query)
    return HttpResponse(Filter(Model, query_dict).query_tags())



# Generated by Django 5.1.7 on 2025-03-16 19:12

import accrete.contrib.user.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('user', '0003_alter_user_email'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='login',
            field=models.CharField(max_length=254, null=True, unique=True, validators=[accrete.contrib.user.models.validate_member_login], verbose_name='Login'),
        ),
        migrations.AlterField(
            model_name='user',
            name='email',
            field=models.EmailField(max_length=254, null=True, unique=True, verbose_name='Email Address'),
        ),
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(condition=models.Q(('email__isnull', False), ('login__isnull', False), _connector='XOR'), name='email_or_login_set', violation_error_message='email or login must be set'),
        ),
    ]



# Generated by Django 5.1.7 on 2025-03-16 18:11

import accrete.contrib.user.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user', '0002_user_theme'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='email',
            field=models.EmailField(max_length=254, unique=True, validators=[accrete.contrib.user.models.validate_member_login], verbose_name='Email Address'),
        ),
    ]



# Generated by Django 5.1.7 on 2025-03-23 10:21

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0005_accessgroup_apply_on_alter_member_access_groups_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name='member',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to=settings.AUTH_USER_MODEL),
        ),
    ]



import re
import ast
import json
from dataclasses import dataclass

from django.core import paginator
from django.db.models import Model
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _
from accrete.contrib.ui import Filter


class Response:

    def __init__(self, *, template: str, context: dict):
        self.template = template
        self.context = context

    @staticmethod
    def add_trigger(response):
        pass

    def get_context(self):
        return self.context

    def render(self, request) -> str:
        return render_to_string(
            template_name=self.template, context=self.get_context(), request=request
        )

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        extra_content = extra_content or ''
        res = HttpResponse(content=(
            self.render(request)
            + render_to_string('ui/message.html', request=request)
            + extra_content
        ))
        self.add_trigger(res)
        if replace_body:
            res.headers['HX-Retarget'] = 'body'
            res.headers['HX-Reswap'] = 'innerHTML'
            res.headers['HX-Push-Url'] = request.path
        return res


class OobResponse(Response):

    oob_template = 'ui/oob.html'

    def __init__(self, *, template: str, context: dict, swap: str, tag: str = 'div'):
        super().__init__(template=self.oob_template, context=context)
        self.include_template = template
        self.swap = swap
        self.tag = tag

    def get_context(self):
        context = super().get_context()
        context.update({'oob': {
            'template': self.include_template,
            'swap': self.swap,
            'tag': self.tag
        }})
        return context


class WindowResponse(Response):

    base_template = 'ui/layout.html'

    def __init__(
        self, *,
        title: str,
        context: dict,
        overview_template: str = None,
        header_template: str = None,
        panel_template: str = None,
        is_centered: bool = False
    ):
        super().__init__(template=self.base_template, context=context)
        self.title = title
        self.overview_template = overview_template
        self.header_template = header_template
        self.panel_template = panel_template
        self.panel_template = panel_template
        self.is_centered = is_centered

    def _has_panel(self):
        return bool(self.panel_template)

    def get_context(self):
        context = super().get_context()
        if 'has_panel' not in context.keys():
            context.update(has_panel=self._has_panel())
        context.update({
            'title': self.title,
            'overview_template': self.overview_template,
            'header_template': self.header_template,
            'panel_template': self.panel_template,
            'is_centered': self.is_centered
        })
        return context

    def response(self, request, extra_content: str = None, replace_body: bool = True) -> HttpResponse:
        return super().response(request, extra_content, replace_body)


class ListResponse(WindowResponse):

    def __init__(
        self, *,
        title: str,
        context: dict,
        list_entry_template: str = None,
        page: paginator.Page = None,
        ui_filter: Filter = None,
        endless_scroll: bool = True,
        header_template: str = None,
        panel_template: str = None,
        column_count: int = 1,
        column_height: str = '150px',
        overview_template: str = 'ui/list.html',
        detail_header_template: str = None,
        detail_data_template: str = None
    ):
        assert page is not None or ui_filter is not None, _(
            'Argument page or ui_filter must be supplied'
        )
        super().__init__(
            title=title,
            context=context,
            overview_template=overview_template,
            header_template=header_template,
            panel_template=panel_template,
        )
        self.page = page or (ui_filter and ui_filter.get_page())
        self.list_entry_template = list_entry_template
        self.ui_filter = ui_filter
        self.endless_scroll = endless_scroll
        self.column_count = column_count
        self.column_height = column_height
        self.detail_header_template = detail_header_template
        self.detail_data_template = detail_data_template

    def _has_panel(self):
        return bool(self.panel_template or self.ui_filter)

    def get_context(self):
        context = super().get_context()
        context.update({
            'list_entry_template': self.list_entry_template,
            'page': self.page,
            'ui_filter': self.ui_filter,
            'endless_scroll': self.endless_scroll,
            'column_count': self.column_count,
            'column_height': self.column_height,
            'detail_header_template': self.detail_header_template,
            'detail_data_template': self.detail_data_template,
            'show_content_right': str(bool(
                self.detail_header_template or self.detail_data_template
            )).lower(),
        })
        return context

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        return super().response(request, extra_content, replace_body)


class ListEntryResponse(Response):

    base_template = 'ui/list_update.html'

    def __init__(
        self, *,
        instance: Model,
        list_entry_template: str,
        context: dict = None,
        page: paginator.Page = None,
        is_new: bool = False,
        column_count: int = 1,
        column_height: str = '150px',
    ):
        super().__init__(template=self.base_template, context=context or {})
        self.instance = instance
        self.list_entry_template = list_entry_template
        self.page = page
        self.is_new = is_new
        self.column_count = column_count
        self.column_height = column_height

    def get_context(self):
        context = super().get_context()
        context.update({
            'instance': self.instance,
            'list_entry_template': self.list_entry_template,
            'is_new': self.is_new,
            'column_count': self.column_count,
            'column_height': self.column_height
        })
        return context

    def render(self, request) -> str:
        res = super().render(request)
        if self.page:
            pagination_update = OobResponse(
                template='ui/layout.html#pagination',
                swap='innerHTML:#pagination',
                context=dict(page=self.page)
            ).render(request)
            res += pagination_update
        return res


class TableResponse(WindowResponse):

    def __init__(
        self, *,
        title: str,
        context: dict,
        instance_label: str,
        fields: list[str],
        footer: dict = None,
        page: paginator.Page = None,
        ui_filter: Filter = None,
        endless_scroll: bool = True,
        header_template: str = None,
        panel_template: str = None,
        overview_template: str = 'ui/table.html',
        detail_header_template: str = None,
        detail_data_template: str = None
    ):
        assert page is not None or ui_filter is not None, _(
            'Argument page or ui_filter must be supplied'
        )
        super().__init__(
            title=title,
            context=context,
            overview_template=overview_template,
            header_template=header_template,
            panel_template=panel_template
        )
        self.instance_label = instance_label
        self.fields = fields
        self.footer = footer
        self.page = page or (ui_filter and ui_filter.get_page())
        self.ui_filter = ui_filter
        self.endless_scroll = endless_scroll
        self.detail_header_template = detail_header_template
        self.detail_data_template = detail_data_template

    def _has_panel(self):
        return bool(self.panel_template or self.ui_filter)

    def get_context(self):
        context = super().get_context()
        context.update({
            'page': self.page,
            'ui_filter': self.ui_filter,
            'endless_scroll': self.endless_scroll,
            'fields': self.fields,
            'instance_label': self.instance_label,
            'footer': self.footer,
            'detail_header_template': self.detail_header_template,
            'detail_data_template': self.detail_data_template,
            'show_content_right': str(bool(
                self.detail_header_template or self.detail_data_template
            )).lower()
        })
        return context

    def response(self, request, extra_content: str = None, replace_body: bool = False) -> HttpResponse:
        return super().response(request, extra_content, replace_body)


class TableRowResponse(Response):

    base_template = 'ui/table_row_update.html'

    def __init__(
        self, *,
        instance: Model,
        fields: list[str],
        footer: dict = None,
        page: paginator.Page = None,
    ):
        super().__init__(template=self.base_template, context={})
        self.instance = instance
        self.fields = fields
        self.footer = footer
        self.page = page

    def get_context(self):
        context = super().get_context()
        context.update({
            'instance': self.instance,
            'fields': self.fields,
            'footer': self.footer,
            'page': self.page
        })
        return context

    def render(self, request) -> str:
        res = super().render(request)
        if self.page:
            pagination_update = OobResponse(
                template='ui/layout.html#pagination',
                swap='innerHTML:#pagination',
                context=dict(page=self.page)
            ).render(request)
            res += pagination_update
        return res


class DetailResponse(Response):

    base_template = 'ui/detail.html'

    def __init__(
        self, *,
        context: dict,
        header_template: str = None,
        data_template: str = None
    ):
        super().__init__(template=self.base_template, context=context)
        self.header_template = header_template
        self.data_template = data_template
        self.context.update()

    @staticmethod
    def add_trigger(response):
        add_trigger(response, 'activate-detail')

    def get_context(self):
        context = super().get_context()
        context.update({
            'detail_header_template': self.header_template,
            'detail_data_template': self.data_template
        })
        return context


class ModalResponse(Response):

    def __init__(
        self, *,
        modal_id: str,
        template: str,
        context: dict,
        title: str = None,
        is_update: bool = False,
        is_blocking: bool = False,
        modal_width: str = None

    ):
        super().__init__(template=template, context=context)
        self.modal_id = modal_id
        self.title = title
        self.is_update = is_update
        self.is_blocking = is_blocking
        self.modal_width = modal_width

    def get_context(self):
        context = super().get_context()
        context.update({
            'title': self.title,
            'modal_id': re.sub(r'[^A-Za-z-]+', '', self.modal_id).strip('-'),
            'is_update': self.is_update,
            'is_blocking': self.is_blocking,
            'modal_width': self.modal_width
        })
        return context


@dataclass
class ClientTrigger:

    trigger: dict | str
    header: str = 'HX-Trigger'


class TriggerResponse:

    def __init__(self, trigger: list[ClientTrigger]):
        self.trigger = trigger

    def response(self):
        res = HttpResponse()
        res.headers['HX-Reswap'] = 'none'
        for trigger in self.trigger:
            add_trigger(res, trigger.trigger, trigger.header)
        return res


def search_select_response(queryset) -> HttpResponse:
    return HttpResponse(render_to_string(
        'ui/widgets/model_search_select_options.html',
        {'options': queryset}
    ))


def message_response(request, persistent: bool = False):
    return HttpResponse(content=(render_to_string(
        'ui/message.html', context={'persistent': persistent}, request=request
    )))


def add_trigger(
    response: HttpResponse,
    trigger: dict | str,
    header: str = 'HX-Trigger'
) -> HttpResponse:
    if isinstance(trigger, str):
        trigger = {trigger: ''}
    res_trigger = response.headers.get(header)
    if not res_trigger:
        response.headers[header] = json.dumps(trigger)
        return response
    try:
        res_trigger = ast.literal_eval(response.headers.get(header, '{}'))
    except SyntaxError:
        res_trigger = {response.headers[header]: ''}
    res_trigger.update(trigger)
    response.headers[header] = json.dumps(res_trigger)
    return response


def update(request, ui_responses: list[Response], trigger: list[ClientTrigger] = None) -> HttpResponse:
    response = HttpResponse()
    content = ''
    for res in ui_responses:
        content += res.render(request)
        res.add_trigger(response)
    content += render_to_string('ui/message.html', request=request)
    response.content = content
    for t in trigger or []:
        add_trigger(response, t.trigger, t.header)
    return response



from .models import Log
from accrete.tenant import get_tenant


def current_log_state(model: str, object_id: int):
    tenant = get_tenant()
    tenant_id = tenant and tenant.id or None
    logs = Log.objects.raw("""
        SELECT
            log_value.id,
            log_value.model,
            log_value.field,
            log_value.object_id,
            log_value.log_date,
            log_value.old_value_type,
            log_value.new_value_type,
            log_value.old_value,
            log_value.new_value
        FROM (
            SELECT model, field, object_id, MAX(log_date) AS log_date
            FROM accrete_log
            WHERE
                model = %s AND
                object_id = %s AND
                tenant_id = %s
            GROUP BY model, field, object_id
        ) log_result
        JOIN accrete_log log_value ON
            log_value.model = log_result.model AND
            log_value.field = log_result.field AND
            log_value.object_id = log_result.object_id AND
            log_result.log_date = log_value.log_date;
    """, [model, object_id, tenant_id])
    return logs



import logging
from django.conf import settings

_logger = logging.getLogger(__name__)

ACCRETE_TENANT_NOT_SET_URL = getattr(
    settings, 'ACCRETE_TENANT_NOT_SET_URL',
    False
)

ACCRETE_GROUP_NOT_SET_URL = getattr(
    settings, 'ACCRETE_GROUP_NOT_SET_URL',
    False
)

if not ACCRETE_TENANT_NOT_SET_URL:
    _logger.warning(
        'Setting ACCRETE_TENANT_NOT_SET_URL missing.'
    )



# Generated by Django 5.1.7 on 2025-03-31 20:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0006_alter_member_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='accessgroup',
            name='description',
            field=models.TextField(blank=True, null=True, verbose_name='Description'),
        ),
    ]



from django.contrib import admin
from . import models


class SystemMailAdmin(admin.ModelAdmin):
    list_display = ('to_addr', 'from_name', 'subject', 'sent', 'error')


admin.site.register(models.SystemMail, SystemMailAdmin)



import re
import logging
from uuid import uuid4
from typing import Type, Any
from dataclasses import dataclass, field
from django.db import transaction
from django.forms import BaseFormSet, Form, ModelForm

_logger = logging.getLogger(__name__)


@dataclass
class FormResult:

    form: Form | ModelForm
    is_saved: bool = False
    save_error: str | None = None
    save_error_id: str | None = None
    res: Any = None  # return value of form.save()
    inline_formsets: list = field(default_factory=list)

    def __getattr__(self, item):
        return getattr(self.form, item)

    def __getitem__(self, item):
        return self.form.__getitem__(item)


def save_form(
    form: Form | ModelForm,
    *,
    commit: bool = True,
    check_changed: bool = False,
    reraise: bool = False
) -> FormResult:
    if not hasattr(form, 'save'):
        raise AttributeError('Form must have method "save" implemented.')
    result = FormResult(form=form)
    changed = not check_changed or form.has_changed()
    try:
        if form.is_valid():
            if changed:
                with transaction.atomic():
                    result.res = form.save(commit=commit)
            result.is_saved = True
    except Exception as e:
        result.save_error = repr(e)
        error_id = str(uuid4())[:8]
        _logger.exception(f'{error_id}: {e}')
        result.save_error_id = error_id
        if reraise:
            raise e
    return result


def save_forms(form, inline_formsets: list = None, commit=True, reraise: bool = False) -> FormResult:

    def handle_error(error):
        result.save_error = repr(error)
        error_id = str(uuid4())[:8]
        _logger.exception(f'{error_id}: {error}')
        result.save_error_id = error_id

    if not hasattr(form, 'save'):
        raise AttributeError('Form must have method "save" implemented.')

    result = FormResult(form=form)
    result.inline_forms = inline_formsets

    try:
        form.is_valid()
        inlines_valid = all([
            inline_formset.is_valid() for inline_formset in result.inline_formsets
        ])
    except Exception as e:
        handle_error(e)
        if reraise:
            raise e
        return result

    if not form.is_valid() or not inlines_valid:
        return result

    try:
        with transaction.atomic():
            result.res = form.save(commit=commit)
            for inline_formset in result.inline_formsets:
                inline_formset.save(commit=commit)
    except Exception as e:
        handle_error(e)
        if reraise:
            raise e
        return result

    result.is_saved = True
    return result


def inline_vals_from_post(post: dict, prefix: str) -> list[dict]:
    post_keys = set(re.findall(f'{prefix}-[0-9]+', ', '.join(post.keys())))
    initial_data = {
        post_key: {}
        for post_key in post_keys if not post.get(f'{post_key}-DELETE')
    }
    for key, val in post.items():
        post_key = '-'.join(key.split('-')[:-1])
        if post_key not in initial_data:
            continue
        field_name = key.split('-')[-1]
        initial_data[post_key].update({field_name: val})
    return [val for val in initial_data.values()]


def extend_formset(formset_class, post: dict, data: list[dict]|dict, **formset_kwargs) -> Type[BaseFormSet]:
    formset = formset_class(post, **formset_kwargs)
    if not formset.is_valid():
        return formset
    form_data = post.copy()
    if isinstance(data, dict):
        data = [data]
    prefix = formset_kwargs.get('prefix', 'form')
    total = int(form_data[f'{prefix}-TOTAL_FORMS']) - 1
    for item in data:
        total += 1
        form_data.update({f'{prefix}-{total}-{key}': value for key, value in item.items()})
    form_data[f'{prefix}-TOTAL_FORMS'] = total + 1
    formset = formset_class(form_data, **formset_kwargs)
    for form in formset:
        form._errors = {}
    return formset



# Generated by Django 5.1.1 on 2025-01-10 07:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('accrete', '0002_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='member',
            name='name',
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _


class SystemMail(models.Model):
    class Meta:
        db_table = 'system_mail'

    from_name = models.CharField(
        verbose_name=_('From'),
        max_length=255
    )

    to_addr = models.TextField(
        verbose_name=_('To'),
        help_text=_('E-Mail addresses, comma separated')
    )

    subject = models.CharField(
        verbose_name=_('Subject'),
        max_length=255,
        default=''
    )

    body = models.TextField(
        verbose_name=_('Mail Body'),
        default=''
    )

    sent = models.BooleanField(
        verbose_name=_('E-Mail sent'),
        default=False
    )

    error = models.TextField(
        verbose_name=_('Errors'),
        null=True,
        blank=True
    )

    def __str__(self):
        return f'{self.to_addr}: {self.subject}'
